from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Union, List
from .columns.column import Column
class Index:
    _name: str = ''
    _index_type: str = ''
    _columns: List[str] = None
    _schema_errors: List[str] = None
    _schema_warnings: List[str] = None
    _parsing_errors: List[str] = None
    _parsing_warnings: List[str] = None

    def __init__(self, name: str = '', columns: List[str] = None, index_type: str = None):
        """ Index constructor

        :param name: The name of the index
        :param columns: The name of the columns in the index
        :param index_type: The type of the index (INDEX, UNIQUE, or PRIMARY)
        """
        self._name = name
        self._columns = [*columns] if columns else []
        self._schema_errors = None
        self._schema_warnings = None
        self._parsing_errors = None
        self._parsing_warnings = None
        # this is often set at the class level by a parser, so we only want to set it if provided
        if index_type is not None:
            self._index_type = index_type

    @property
    def name(self) -> str:
        """ Public getter.  Returns the name of the column.

        :returns: The index name
        """
        # if we don't have a name then assume it's name matches the name of our first column
        if self._name:
            return self._name
        if len(self._columns) > 0:
            return self._columns[0] if type(self._columns[0]) == str else self._columns[0].name
        return ''

    @property
    def index_type(self) -> str:
        """ Public getter.  Returns a string denoting the type of the index.  Always returns in uppercase

        Index type can be 'INDEX', 'UNIQUE', or 'PRIMARY'

        :returns: The index type
        """
        return self._index_type.upper()

    def is_primary(self) -> bool:
        """ Returns True/False to denote if the index is a primary index """
        return self.index_type == 'PRIMARY'

    @property
    def columns(self) -> List[Column]:
        """ Returns a list of the columns on the index. """
        return self._columns

    @property
    def schema_errors(self) -> List[str]:
        """ Returns a list of schema errors """
        if self._schema_errors is None:
            self._check_for_schema_errors_and_warnings()
        return self._schema_errors

    @property
    def schema_warnings(self) -> List[str]:
        """ Returns a list of schema warnings """
        if self._schema_errors is None:
            self._check_for_schema_errors_and_warnings()
        return self._schema_warnings

    @property
    def parsing_errors(self) -> List[str]:
        """ Returns a list of parsing errors """
        return self._parsing_errors if self._parsing_errors is not None else []

    @property
    def parsing_warnings(self) -> List[str]:
        """ Returns a list of parsing/table warnings """
        return self._parsing_warnings if self._parsing_warnings is not None else []

    def _check_for_schema_errors_and_warnings(self):
        self._schema_errors = []
        self._schema_warnings = []
        for required in ['name', 'index_type', 'columns']:
            if not getattr(self, required):
                self._schema_errors.append(f"Missing {required} for index {self.name}")
        if len(self.name) > 64:
            self._schema_errors.append('Key name %s must be <=64 characters long' % (self._name))

    def __str__(self):
        """ Returns the MySQL command that would create the index

        i.e. PRIMARY KEY `id` (`id`)'

        :returns: A partial MySQL command that could be used to generate the column
        :rtype: string
        """
        parts = []
        if self.index_type == 'PRIMARY':
            parts.append('PRIMARY')
        elif self.index_type == 'UNIQUE':
            parts.append('UNIQUE')
        parts.append('KEY')

        if self._name:
            parts.append('`%s`' % self.name)
        parts.append('(`%s`)' % ("`,`".join(self.columns)))

        return ' '.join(parts)



from .base import base
import mygrations
def execute(options, print_results=True):

    obj = Version(options)
    return obj.execute(print_results=print_results)
class Version(base):
    def execute(self, print_results=True):
        version_info = [
            f'mygrations v{mygrations.__version__}',
            'Copyright (c) 2022 Conor Mancone (cmancone@gmail.com)',
            'MIT License',
        ]
        if print_results:
            "\n".join(version_info)
        return (print_results, True)



import unittest
from .remove_constraint import RemoveConstraint
from ..definitions.constraint import Constraint
class RemoveConstraintTest(unittest.TestCase):
    def test_as_string(self):
        constraint = Constraint('user_id_fk', 'user_id', 'users', 'id')
        self.assertEquals("DROP FOREIGN KEY `user_id_fk`", str(RemoveConstraint(constraint)))



from collections import OrderedDict

from mygrations.core.parse.parser import Parser
from mygrations.formats.mysql.definitions.constraint import Constraint
from mygrations.formats.mysql.definitions.index import Index
from mygrations.formats.mysql.definitions.table import Table

from .parsers import *
class CreateParser(Table, Parser):

    # this defines the rules for the parsing engine.  Yes, I decided to try to build
    # a parsing engine to parse the MySQL.  Seems like a reasonable choice at the
    # moment, but we'll see how it works out in the long run.  The setup you see here
    # is a compromise between the declarative programming principles I like and the
    # time it takes to actually build such a system, especially for a relatively
    # limited use-case.  In the long run if this works well it could probably
    # become its own multi-purpose tool.  These "rules" basically boil down to:
    #
    # CREATE TABLE table_name ( [definitions] ) [options];
    #
    # definitions and options have their own patterns of allowed values and rules,
    # which also have to be defined.  That is handled below for the sake of brevity
    rules = [{
        'type': 'literal',
        'value': 'CREATE TABLE'
    }, {
        'type': 'regexp',
        'value': '\S+',
        'name': 'name'
    }, {
        'type': 'literal',
        'value': '('
    }, {
        'type':
        'children',
        'name':
        'definitions',
        'classes': [
            IndexPrimary, IndexKey, IndexUnique, ConstraintForeign, TypeCharacter, TypeNumeric, TypeDecimal, TypeText,
            TypeEnum, TypePlain, TypeDateTime
        ]
    }, {
        'type': 'literal',
        'value': ')'
    }, {
        'type': 'children',
        'name': 'table_options',
        'classes': [TableOption],
        'optional': True
    }, {
        'type': 'literal',
        'value': ';',
        'optional': True,
        'name': 'closing_semicolon'
    }]

    def process(self):

        self.semicolon = True if 'closing_semicolon' in self._values else False
        self._name = self._values['name'].strip('`')
        self._definitions = self._values['definitions']
        self._options = self._values['table_options'] if 'table_options' in self._values else []
        self._columns = OrderedDict()
        self._indexes = OrderedDict()
        self._constraints = OrderedDict()
        self._primary = ''

        # ignore the AUTO_INCREMENT option: there is no reason for us to ever manage that
        self._options = [opt for opt in self._options if opt.name != 'AUTO_INCREMENT']

        for definition in self._definitions:
            if hasattr(definition, 'as_definition'):
                if definition._name in self._columns:
                    self._global_errors.append(f"Table '{self._name}' has two columns named '{definition._name}'")
                    continue

                self.add_column(definition.as_definition())    # see notes on definition.as_definition()
            elif isinstance(definition, Index):
                if definition._name in self._indexes:
                    self._global_errors.append(f"Table '{self._name}' has two indexes named '{definition._name}'")
                    continue

                self.add_index(definition)
            elif isinstance(definition, Constraint):
                if definition._name in self._constraints:
                    self._global_errors.append(f"Table '{self._name}' has two constraints named '{definition._name}'")
                    continue

                self.add_constraint(definition)

            else:
                raise ValueError("Found unknown definition of type ".definition.__class__)

        if not self.semicolon:
            self._global_errors.append("Missing ending semicolon for table %s" % self._name)



import pymysql as pymysql_driver    # alias to avoid confusion
from collections import OrderedDict
class PyMySQL(object):
    """ High level driver for a MySQLdb connection """
    execute_cursor = None

    def __init__(self, credentials=None, connection=None):
        """ Initialize the PyMySQL connection

        Accepts either a dictionary with mysql connnection credentials or a PyMySQL connection object.
        """
        if credentials is not None:
            self.connection = pymysql_driver.connect(
                user=credentials['user'],
                password=credentials['password'],
                host=credentials['host'],
                database=credentials['database'],
                autocommit=False,
                connect_timeout=2,
                cursorclass=pymysql_driver.cursors.DictCursor
            )
        elif connection is not None:
            self.connection = connection
        else:
            raise ValueError("Must provide either the database credentials or connection object")

    def tables(self):
        """ Returns the tables in the connected database

        :returns: An ordered dict with table definitions by table name
        :rtype: OrderedDict
        """
        cursor = self.connection.cursor()

        # not returning an iterator: just fetch everything.
        # I'm guessing this will be fine for any realistic database
        # size, and avoids issues of having multiple open cursors
        # at the same time.
        cursor.execute('SHOW TABLES')
        table_names = []
        for result_data in cursor:
            for table_name in result_data.values():
                table_names.append(table_name)

        definitions = OrderedDict()
        for table_name in table_names:
            cursor.execute('SHOW CREATE TABLE %s' % table_name)
            if not cursor.rowcount:
                raise ValueError("Failed to execute SHOW CREATE TABLE command on table %s" % table_name)

            result_data = cursor.fetchone()
            definitions[table_name] = result_data['Create Table']

        cursor.close()

        return definitions

    def rows(self, table_name):
        """ Returns the rows in the table as a tuple of dicts

        :returns: The rows in the table
        :rtype: (dict)
        """

        # still sticking to full caching client-side without
        # support for memory-preserving iterators.  Given our
        # use case, I think this will be fine.  Can always change later
        cursor = self.connection.cursor()
        cursor.execute('SELECT * FROM %s' % table_name)
        rows = [row for row in cursor]
        cursor.close()

        return rows

    def execute(self, queries):
        if isinstance(queries, str):
            queries = [queries]

        cursor = self.connection.cursor()
        for query in queries:
            cursor.execute(query)
        self.connection.commit()
        cursor.close()



from .base import Base
from mygrations.formats.mysql.file_reader.database import Database as DatabaseParser
from mygrations.formats.mysql.db_reader.database import Database as DatabaseReader
from mygrations.formats.mysql.mygrations.mygration import Mygration
from mygrations.formats.mysql.mygrations.row_mygration import RowMygration
from mygrations.formats.mysql.mygrations.operations.disable_checks import DisableChecks
from mygrations.formats.mysql.mygrations.operations.enable_checks import EnableChecks
import sys
def execute(options, print_results=True):

    obj = Plan(options)
    return obj.execute(print_results=print_results)
class Plan(Base):
    needs_cursor = True

    def execute(self, print_results=True):

        commands = ['%s' % command for command in self.build_commands()]
        if print_results:
            print("\n".join(commands))
        return (commands, True)

    def build_commands(self):

        files_database = DatabaseParser(self.get_sql_files())

        # any errors or warnings?
        quit_early = False
        if files_database.errors and not self.options.get('force'):
            print('Errors found in *.sql files', file=sys.stderr)
            quit_early = True
            for error in files_database.errors:
                print(error, file=sys.stderr)

        if quit_early:
            return []

        # use the credentials to load up a database connection
        live_database = DatabaseReader(self.get_driver())

        # we have to tell the live database to load records
        # for any tables we are tracking records for.
        # We use the "files" database as the "truth" because
        # just about any database can have records, but
        # that doesn't mean we want to track them.  We only
        # track them if the file has rows.
        for table in files_database.tables.values():
            if not table.tracking_rows:
                continue

            if not table.name in live_database.tables:
                continue

            live_database.read_rows(table)

        mygrate = Mygration(files_database, live_database, False)

        ops = []
        if mygrate.operations:
            ops.extend(mygrate.operations)

        rows = RowMygration(files_database, live_database)
        if rows.operations:
            ops.extend(rows.operations)

        if not ops:
            return []

        return [
            DisableChecks(),
            *ops,
            EnableChecks(),
        ]



from mygrations.core.definitions.constraint import Constraint as BaseConstraint
class Constraint(BaseConstraint):
    def __init__(
        self, name='', column_name='', foreign_table_name='', foreign_column_name='', on_delete='', on_update=''
    ):
        super().__init__(
            name=name,
            column_name=column_name,
            foreign_table_name=foreign_table_name,
            foreign_column_name=foreign_column_name,
            on_delete=on_delete,
            on_update=on_update
        )



import unittest

from mygrations.formats.mysql.file_reader.parsers.type_plain import TypePlain
class TestTypePlain(unittest.TestCase):
    def test_simple(self):

        # parse typical insert values
        parser = TypePlain()
        returned = parser.parse("created date NOT NULL DEFAULT 'bob',")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('created', parser._name)
        self.assertEquals('date', parser._column_type)
        self.assertFalse(parser._null)
        self.assertEquals('bob', parser._default)
        self.assertTrue(parser.has_comma)
        self.assertEquals(0, len(parser._parsing_errors))
        self.assertEquals(0, len(parser._parsing_warnings))

    def test_optional_default(self):

        # parse typical insert values
        parser = TypePlain()
        returned = parser.parse("created date")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('created', parser._name)
        self.assertEquals('date', parser._column_type)
        self.assertTrue(parser._null)
        self.assertEquals(None, parser._default)
        self.assertFalse(parser.has_comma)
        self.assertEquals(0, len(parser._parsing_errors))
        self.assertEquals(0, len(parser._parsing_warnings))

    def test_strip_backticks(self):

        # parse typical insert values
        parser = TypePlain()
        returned = parser.parse("`created` date")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)
        self.assertEquals('created', parser._name)



from mygrations.core.definitions.columns.string import String as StringBase
from typing import List, Union
class String(StringBase):
    def __init__(
        self,
        name: str = '',
        column_type: str = '',
        length: Union[str, int] = None,
        null: bool = True,
        has_default: bool = False,
        default: Union[str, int] = None,
        unsigned: bool = None,
        character_set: str = None,
        collate: str = None,
        auto_increment: bool = False,
        enum_values: List[str] = None,
        parsing_errors: List[str] = None,
        parsing_warnings: List[str] = None,
    ):
        # it would be nice to just do `def __init__(**kwargs)` and then `super().__init__(**kwargs)`
        # but then we would lose our type hints.  :shrug:
        super().__init__(
            name=name,
            column_type=column_type,
            length=length,
            null=null,
            has_default=has_default,
            default=default,
            character_set=character_set,
            collate=collate,
            auto_increment=auto_increment,
            enum_values=enum_values,
            parsing_errors=parsing_errors,
            parsing_warnings=parsing_warnings,
            unsigned=unsigned,
        )



import unittest
from .string import String
class TestString(unittest.TestCase):
    def test_string_conversion_with_collation(self):
        column = String('test_column', 'VARCHAR', length=25, default='asdf', character_set='unicode', collate='sup')

        self.assertEquals(
            '`test_column` VARCHAR(25) DEFAULT \'asdf\' CHARACTER SET \'UNICODE\' COLLATE \'SUP\'', str(column)
        )
        self.assertEquals([], column.schema_errors)
        self.assertEquals([], column.schema_warnings)

    def test_is_the_same(self):
        attrs = {
            'length': 255,
            'character_set': 'unicode',
        }

        column1 = String('test', 'VARCHAR', **attrs)
        column2 = String('test', 'VARCHAR', **{**attrs, 'collate': 'hey'})

        # since collate is only present for one they are the same
        self.assertTrue(column1.is_really_the_same_as(column2))
        self.assertTrue(column2.is_really_the_same_as(column1))

        # but since they both have values but are different they are different
        column3 = String('test', 'VARCHAR', **{**attrs, 'collate': 'sup'})
        self.assertFalse(column3.is_really_the_same_as(column2))
        self.assertFalse(column2.is_really_the_same_as(column3))

    def test_no_default(self):
        no_default = [
            'TINYBLOB',
            'BLOB',
            'MEDIUMBLOB',
            'LONGBLOB',
            'TINYTEXT',
            'TEXT',
            'MEDIUMTEXT',
            'LONGTEXT',
            'JSON',
        ]
        for column_type in no_default:
            column = String('test', column_type, default='Hey', has_default=True)
            self.assertEquals(
                [f"Column 'test' of type '{column_type}' cannot have a default"],
                column.schema_errors,
            )

    def test_no_character_set_or_collate(self):
        nope = [
            'TINYBLOB',
            'BLOB',
            'MEDIUMBLOB',
            'LONGBLOB',
            'TINYTEXT',
            'TEXT',
            'MEDIUMTEXT',
            'LONGTEXT',
            'JSON',
        ]
        for column_type in nope:
            column = String('test', column_type, character_set='Hey')
            self.assertEquals(
                [f"Column 'test' of type '{column_type}' cannot have a collation/character set"],
                column.schema_errors,
            )
            column = String('test', column_type, collate='Hey')
            self.assertEquals(
                [f"Column 'test' of type '{column_type}' cannot have a collation/character set"],
                column.schema_errors,
            )

    def test_no_length(self):
        no_length = [
            'TINYBLOB',
            'BLOB',
            'MEDIUMBLOB',
            'LONGBLOB',
            'TINYTEXT',
            'TEXT',
            'MEDIUMTEXT',
            'LONGTEXT',
            'JSON',
        ]
        for column_type in no_length:
            column = String('test', column_type, length='28')
            self.assertEquals(
                [f'Column test of type {column_type} cannot have a length'],
                column.schema_errors,
            )

    def test_binary_collation(self):
        binary_only = [
            'BINARY',
            'VARBINARY',
        ]
        for column_type in binary_only:
            column = String('test', column_type, character_set='Hey')
            self.assertEquals(
                [f"Column 'test' of type '{column_type}' can only have a collate/character set of BINARY"],
                column.schema_errors,
            )
            column = String('test', column_type, collate='Hey')
            self.assertEquals(
                [f"Column 'test' of type '{column_type}' can only have a collate/character set of BINARY"],
                column.schema_errors,
            )
            column = String('test', column_type, collate='binary')
            self.assertEquals([], column.schema_errors)
            column = String('test', column_type, character_set='binary')
            self.assertEquals([], column.schema_errors)
            column = String('test', column_type, character_set='binary', collate='binary')
            self.assertEquals([], column.schema_errors)

    def test_no_auto_increment(self):
        column = String('test', 'VARCHAR', auto_increment=True)
        self.assertEquals(
            [f"Column 'test' of type 'VARCHAR' cannot be an AUTO_INCREMENT: only numeric columns can"],
            column.schema_errors,
        )
